/*********************************************************************
 *
 * launcher.c - an app to launch apps
 *
 * Author: Ammon Riley <ammon@sbcglobal.net>
 *
 *********************************************************************/

#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>	
#include <sys/fcntl.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include "asm/arch/hijack.h"

/*********************************************************************
 * The menu is implemented as a circularly linked list. It goes
 * forward and backward, as well as up and down. The only menu
 * entries that can go down are menus. All entries know the menu
 * that contains them. By virtue of the config.ini file syntax,
 * there cannot be a menu without at least one entry in it.
 *********************************************************************/
typedef struct MenuItem {
    char *name;
    char *path;			/* If NULL, node is a container. */
    char *args;
    struct MenuItem *contents;	/* NULL if path != NULL */
    struct MenuItem *parent;
    struct MenuItem *next;
    struct MenuItem *prev;
} MenuItem;

static struct MenuItem *root_menu = NULL;

/*********************************************************************
 * A few error codes for problems we might run into while parsing
 * the config.ini and building the menu.
 *********************************************************************/
#define BUILD_OK			0x01
#define BUILD_OPEN_ERR			0x02
#define BUILD_NO_CONFIG			0x03
#define BUILD_MEM_ERR			0x04

/*********************************************************************
 * The very last item that will be added to the menu system is
 * an option to rebuild the menus, so you can add applications
 * and edit your config.ini file without the need to restart 
 * launcher. 
 *********************************************************************/
#define REBUILD_TEXT	"[Rebuild menu]"

/*********************************************************************
 * The UI buttons we'll be using. Should be pretty intuitive.
 *********************************************************************/
static  unsigned long buttons[] = {
    20,
    IR_RIO_NEXTTRACK_PRESSED,
    IR_RIO_PREVTRACK_PRESSED,
    IR_RIO_MENU_PRESSED,
    IR_RIO_CANCEL_PRESSED,
    IR_TOP_BUTTON_PRESSED,
    IR_RIO_VOLMINUS_PRESSED,
    IR_RIO_VOLPLUS_PRESSED, 
    IR_RIGHT_BUTTON_PRESSED,
    IR_LEFT_BUTTON_PRESSED,
    IR_BOTTOM_BUTTON_PRESSED,
    IR_KNOB_RIGHT,
    IR_KNOB_LEFT,
    IR_KNOB_PRESSED,

    /* Kenwood Remote */
    IR_KW_DNPP_PRESSED,
    IR_KW_NEXTTRACK_PRESSED,
    IR_KW_PREVTRACK_PRESSED,
    IR_KW_VOL_PLUS_PRESSED,
    IR_KW_VOL_MINUS_PRESSED,
    IR_KW_STAR_PRESSED
};

/*********************************************************************
 * The amount of the screen we'll be using. I think.
 *********************************************************************/
static hijack_geom_t fullscreen = {
    0, EMPEG_SCREEN_ROWS - 1,
    0, EMPEG_SCREEN_COLS - 1
};

/*********************************************************************
 * destroy_launcher_menu()
 *
 * On rebuilding the menu, we need to free up all the memory
 * already allocated to the menu. This is a nice recursive
 * function to do that.
 *
 * Save the whales and the empeg memory.
 *
 *********************************************************************/
void destroy_launcher_menu(struct MenuItem *item)
{
    if (item != NULL)
	return;			/* Not much to free here. */

    /* Disconnect from the previous node (to prevent a circle). */
    if (item->prev != NULL) {
	item->prev->next = NULL;
	item->prev = NULL;
    }

    /* Delete the next node before losing track of it. */
    if (item->next != NULL) {
	destroy_launcher_menu(item->next);
	item->next = NULL;
    }

    /* Free up submenus. */
    if (item->contents != NULL) {
	destroy_launcher_menu(item->contents);
	item->contents = NULL;
    }

    /* Finally, we can commit suicide. */
    if (item->name != NULL)
	free(item->name);
    if (item->path != NULL)
	free(item->path);
    if (item->args != NULL)
	free(item->args);
    free(item);

    item = NULL;
}


/*********************************************************************
 * get_submenu()
 *
 * Finds the specified submenu from the current menu's contents.
 * If menu_path says there is supposed to be a menu, but get_submenu()
 * can't find it, then it will get built automatically. This feature
 * keeps the config.ini syntax much simpler.
 *
 * Returns NULL on error. Expects a non-null menu node passed in.
 *********************************************************************/
struct MenuItem* get_submenu(struct MenuItem *i_menu, char *menu_path)
{
    struct MenuItem* m_item;
    struct MenuItem* r_item;
    char *next_dot;

    /* Sane input? */
    if (i_menu == NULL || menu_path == NULL)
	return NULL;

    /* Get the first component of the menu hierarchy. */
    next_dot = index(menu_path, '.');
    if (next_dot != NULL)
	next_dot[0] = '\0';

    if (i_menu->contents == NULL) {

	/* Current menu has no entries. We have to add one. */
	m_item = calloc(1, sizeof(struct MenuItem));
	if (m_item == NULL)
	    return NULL;

	m_item->name = strdup(menu_path);
	m_item->parent = i_menu;
	m_item->next = m_item;
	m_item->prev = m_item;

	i_menu->contents = m_item;
	i_menu = m_item;
    }
    else {

	/* Current menu has contents. We'll have to search through
	 * to see if the one we want is already there. */

	/* Break the forward circle temporarily so we don't loop
	 * and loop and loop and loop and... */
	r_item = i_menu->contents;
	r_item->prev->next = NULL;

	while (r_item != NULL) {
	    /* Stop if this is the menu. */
	    if (r_item->path == NULL &&
		strcmp(r_item->name, menu_path) == 0) {
		break;
	    }

	    /* This is not the menu you are looking for... */
	    r_item = r_item->next;
	}

	/* Let the circle be unbroken. */
	i_menu->contents->prev->next = i_menu->contents;

	if (r_item != NULL) {

	    /* Jump into next menu. */
	    i_menu = r_item;
	}
	else {

	    /* No matching menu was found. Gotta make one. */
	    m_item = calloc(1, sizeof(struct MenuItem));
	    if (m_item == NULL)
		return NULL;

	    m_item->name = strdup(menu_path);
	    m_item->parent = i_menu;

	    /* Attach this as the last entry. */
	    m_item->next = i_menu->contents;
	    m_item->prev = i_menu->contents->prev;
	    i_menu->contents->prev->next = m_item;
	    i_menu->contents->prev = m_item;

	    /* Jump into the new menu. */
	    i_menu = m_item;
	}
    }

    /* If there are more components in the path, then deal with them. */
    if (next_dot != NULL) {
	return get_submenu(i_menu, next_dot + sizeof(char));
    }
    else {
	return i_menu;
    }
}

/*********************************************************************
 * build_launcher_menu()
 *
 * This does all the work of parsing the config.ini file, and 
 * creating the menu structure. No surprise there, eh?
 *
 * The syntax that this expects:
 * 	approot=<path>
 * 	item=[menu[.menu[...]].]<Description>,<path to binary>[,args]
 *
 * Anything else is ignored. If <path to binary> is not a fully
 * qualified path, then approot will be appended to it. By virtue 
 * of having the application specify the menu hierarchy it lives
 * in, we are guaranteed that there will never be a menu without at
 * least one entry.
 *********************************************************************/
int build_launcher_menu()
{
    FILE *filedes;
    int i = 0;
    int in_section = 0;
    char *approot = NULL;
    struct MenuItem *i_menu;	/* Pointer to menu to insert item in. */
    struct MenuItem *r_item;	/* Used as temp pointer when inserting. */
    struct MenuItem *item;	/* A brand new menu entry. */
    char *menu_path;
    char buffer[128];
    
    filedes = fopen("/drive0/var/config.ini", "r");
    if (filedes == NULL) 
	return BUILD_OPEN_ERR;

    /* Skip up to our section of the config.ini */
    while (fgets(buffer, 128, filedes) != NULL) {
	/* Hack off any \r, \n, \t and " ". */
	for (i = strlen(buffer) - 1;
		(i >= 0) && (buffer[i] == '\r' ||
		    	     buffer[i] == '\n' ||
			     buffer[i] == '\t' ||
			     buffer[i] == ' ');
		i--) {
	    buffer[i] = '\0';
	}

	if (strcmp(buffer, "[launcher]") == 0) {
	    in_section = 1;
	    break;
	}
    }

    /* Make sure we are actually in our section, and not EOF. */
    if (! in_section) {
	fclose(filedes);
	return BUILD_NO_CONFIG;
    }

    /* Create the root menu. */
    if (root_menu == NULL) {
	root_menu = calloc(1, sizeof(struct MenuItem));
	if (root_menu == NULL) {
	fclose(filedes);
	return BUILD_MEM_ERR;
	}
    }
    root_menu->name = strdup("Root Menu");
    root_menu->path = NULL;
    root_menu->args = NULL;
    root_menu->contents = NULL;
    root_menu->parent = root_menu;
    root_menu->next = NULL;
    root_menu->prev = NULL;

    /* Process our section. */
    while (fgets(buffer, 128, filedes) != NULL) {

	/* We can quit if we reach someone else's section. */
	if (buffer[0] == '[') {
	    fclose(filedes);
	    break;
	}

	/* Hack off any \r, \n, \t and " ". */
	for (i = strlen(buffer) - 1;
		(i >= 0) && (buffer[i] == '\r' ||
		    	     buffer[i] == '\n' ||
			     buffer[i] == '\t' ||
			     buffer[i] == ' ');
		i--) {
	    buffer[i] = '\0';
	}
	
	/* Only process things we understand -- item and approot. */
	if (strncmp(buffer, "item=", 5) == 0) {
	    char *menu_text = buffer + sizeof(char) * 5;
	    char *cmdline = index(buffer, ',');
	    char *args = NULL;

	    /* Is there a command line present? */
	    if (cmdline != NULL) {
		cmdline[0] = '\0';
		cmdline += sizeof(char);

		args = index(cmdline, ',');
		if (args != NULL) {
		    args[0] = '\0';
		    args += sizeof(char);
		}
	    }

	    /* Did we get an actual menu item? */
	    if (strlen(menu_text) && strlen(cmdline)) {

		/* Figure out if this goes in a submenu. */
		char *last_dot = rindex(menu_text, '.');
		i_menu = root_menu;
		if (last_dot != NULL) {
		    menu_path = menu_text;
		    menu_text = last_dot + sizeof(char);
		    last_dot[0] = '\0';

		    /* Get the submenu to drop this item into. */
		    i_menu = get_submenu(i_menu, menu_path);
		    if (i_menu == NULL) {
			fclose(filedes);
			return BUILD_MEM_ERR;
		    }
		}

		/* Create the menu item. */
		item = calloc(1, sizeof(struct MenuItem));
		if (item == NULL) {
		    fclose(filedes);
		    return BUILD_MEM_ERR;
		}

		item->name = strdup(menu_text);
		if (item->name == NULL) {
		    free(item);
		    fclose(filedes);
		    return BUILD_MEM_ERR;
		}

		/* Prepend the approot path, if necessary. */
		if ((approot != NULL) && (cmdline[0] != '/')) {
		    item->path = calloc(strlen(cmdline) + strlen(approot) + 2,
			    		sizeof(char));
		    if (item->path == NULL) {
			free(item->name);
			free(item);
			fclose(filedes);
			return BUILD_MEM_ERR;
		    }
		    strcpy(item->path, approot);
		    strcat(item->path, "/");
		    strcat(item->path, cmdline);
		}
		else {
		    item->path = strdup(cmdline);
		    if (item->path == NULL) {
			free(item->name);
			free(item);
			fclose(filedes);
			return BUILD_MEM_ERR;
		    }
		}

		if (args != NULL) {
		    item->args = strdup(args);
		    if (item->args == NULL) {
			free(item->name);
			free(item->path);
			free(item);
			fclose(filedes);
			return BUILD_MEM_ERR;
		    }
		}
		else {
		    item->args = NULL;
		}

		/* Non-menus never have contents. */
		item->contents = NULL;

		/* Add the new menu item. */
		if (i_menu->contents == NULL) {
		    i_menu->contents = item;
		    item->parent = i_menu;

		    /* Since it's only item, egotisticaly point to self. */
		    item->next = item;
		    item->prev = item;
		}
		else {

		    /* Get the first item in the menu contents, and 
		     * add the new item in front of it -- i.e. at the
		     * end of the list. */
		    r_item = i_menu->contents;
		    item->parent = i_menu;

		    /* Get the new item pointing to the right place. */
		    item->next = r_item;
		    item->prev = r_item->prev;

		    /* Make sure a couple things point to new item. */
		    item->prev->next = item;
		    r_item->prev = item;
		}
	    }
	}
	else if (strncmp(buffer, "approot=", 8) == 0) {
	    if (approot != NULL) 
		free(approot);

	    approot = strdup(buffer + sizeof(char) * 8); 
	    if (approot == NULL)
		return BUILD_MEM_ERR;
	}
    }

    fclose(filedes);
   
    /* Add the rebuild menu option. */
    item = calloc(1, sizeof(struct MenuItem));
    if (item == NULL)
	return BUILD_MEM_ERR;

    item->name = strdup(REBUILD_TEXT);
    if (item->name == NULL) {
	free(item);
	return BUILD_MEM_ERR;
    }
    item->parent = root_menu;

    /* Add the new menu item. */
    if (root_menu->contents == NULL) {
	root_menu->contents = item;
	item->next = item;
	item->prev = item;
    }
    else {
	r_item = root_menu->contents; // Should always be last on root menu
	item->next = r_item;
	item->prev = r_item->prev;
	item->prev->next = item;
	r_item->prev = item;
    }

    return BUILD_OK;
}

/*********************************************************************
 * launch()
 *
 * This runs the application!
 *********************************************************************/
void launch(int fd, struct MenuItem *item)
{
    unsigned long button_press;
    int pid = fork();
    switch (pid) {
	case -1 : /* Failed fork! */
	    ioctl( fd, EMPEG_HIJACK_DISPCLEAR, 0 );
	    ioctl( fd, EMPEG_HIJACK_DISPTEXT, "\n\nUnable to run program!" );
	    ioctl( fd, EMPEG_HIJACK_WAITBUTTONS, &button_press );  
	    break;

	case 0 : /* Child process. */
	    /* Unbind all the parent's stuff. The child will rebind what
	     * it needs, if it needs anything. */
	    ioctl( fd, EMPEG_HIJACK_WAITMENU, NULL );
	    close( fd );
	    execlp(item->path, item->path, item->args, NULL);
	    break; /* Shouldn't need this... */

	default : /* Parent. */
	    waitpid(pid, NULL, 0);
    }
}

/*********************************************************************
 * Viewer main function
 *********************************************************************/
void show_launcher( int fd, hijack_geom_t fullscreen )
{
    unsigned long button_press;
    struct MenuItem *top_item; /* Top item on the display */
    struct MenuItem *cur_item; /* Current item we're adding to the display. */

    /* Bind buttons and grab the screen. I'm trying to bind and unbind
     * these as tightly as possible, because it's possible that having
     * them bound unnecessarily causes conflicts with other running
     * apps. Just a theory, but it can't hurt. */
    ioctl( fd, EMPEG_HIJACK_BINDBUTTONS, buttons );
    ioctl( fd, EMPEG_HIJACK_SETGEOM, &fullscreen );

build_menu_label:	/* Yep, a label. I'm using a goto. */

    if ((root_menu == NULL) && (build_launcher_menu() != BUILD_OK)) {
	ioctl( fd, EMPEG_HIJACK_DISPCLEAR, 0 );
	ioctl( fd, EMPEG_HIJACK_DISPTEXT, "\n\nUnable to build menu!" );
	ioctl( fd, EMPEG_HIJACK_WAITBUTTONS, &button_press );  
	return;
    }

    top_item = root_menu->contents;

    /* Keep scrolling through the menus until the user cancels. */
    while ( 1 )
    {
	char *display_text;	/* The complete display text. */
	char *display_items[4]; /* The individual display items. */
	int j;			/* Just a loop counter. */
	int display_len = 0;	/* A tally for the length of display_text. */

	/* Continue the scroll from the top item. */
	cur_item = top_item;

	/* Grab 4 items -- the max we can fit on the display. */
	for (j = 0; j < 4; j++) {
	    display_items[j] = cur_item->name;

	    display_len += strlen(cur_item->name) + 1; /* To alloc enough. */

	    cur_item = cur_item->next;

	    /* Stop if we've looped back around to the first item -- there
	     * aren't four items in the menu. */
	    if (cur_item == top_item) {
		j++;
		break;
	    }
	}

	/* If there weren't four items, blank out the extra lines. */
	for (;j < 4; j++) {
	    display_items[j] = NULL;
	}

	/* Assemble the display text. */
	display_text = calloc(display_len + 1, sizeof(char));
	strcpy(display_text, display_items[0]); /* There's always one! */
	for (j = 1; j < 4; j++) {
	    strcat(display_text, "\n");
	    if (display_items[j] != '\0') /* SEGV if strcat(str, NULL)! */
		strcat(display_text, display_items[j]);
	}

	/* Clear the previous display, and display next screen. */
	ioctl( fd, EMPEG_HIJACK_DISPCLEAR, 0 );
	ioctl( fd, EMPEG_HIJACK_DISPTEXT, display_text );

	/* Free this up now, in case user decides to exit. */
	free(display_text);
	display_text = NULL;

	/* Wait for the next bit of input */
	ioctl( fd, EMPEG_HIJACK_WAITBUTTONS, &button_press );  

	switch (button_press) {
	    case IR_KNOB_RIGHT :		/* Next menu entry. */
	    case IR_RIGHT_BUTTON_PRESSED :
	    case IR_RIO_NEXTTRACK_PRESSED :
	    case IR_RIO_VOLMINUS_PRESSED :
	    case IR_KW_NEXTTRACK_PRESSED :
	    case IR_KW_VOL_MINUS_PRESSED :
		top_item = top_item->next;
		break;

	    case IR_KNOB_LEFT :			/* Previous menu entry. */
	    case IR_LEFT_BUTTON_PRESSED :
	    case IR_RIO_PREVTRACK_PRESSED :
	    case IR_RIO_VOLPLUS_PRESSED :
	    case IR_KW_PREVTRACK_PRESSED :
	    case IR_KW_VOL_PLUS_PRESSED  :
		top_item = top_item->prev;
		break;

	    case IR_KNOB_PRESSED :		/* Lights, camera, action! */
	    case IR_RIO_MENU_PRESSED :
	    case IR_BOTTOM_BUTTON_PRESSED :
	    case IR_KW_DNPP_PRESSED :
		/* Rebuild menu? */
		if (strcmp(top_item->name, REBUILD_TEXT) == 0) {
		    destroy_launcher_menu(root_menu);
		    root_menu = NULL;

		    /* Break out of the switch() and the while(), all
		     * in one fell swoop. Don't ya love the goto? */
		    goto build_menu_label;
		}

		/* Drop into a submenu? */
		else if ((top_item->path == NULL) &&
			 (top_item->args == NULL)) {
		    if (top_item->contents != NULL) /* Just in case... */
			top_item = top_item->contents;
		    break;
		}

		/* Run a program? */
		else {
		    launch(fd, top_item);

		    /* Only the parent should get here... */
		    return;
		}

	    case IR_RIO_CANCEL_PRESSED :	/* Move up a menu or quit. */
	    case IR_TOP_BUTTON_PRESSED :
	    case IR_KW_STAR_PRESSED :
		/* Only quit the app if we're at the root level. */
		if (top_item->parent != root_menu) {
		    top_item = top_item->parent;
		    break;
		}
		else {
		    /* Leaving our program... unbind everything we bound. */
		    ioctl( fd, EMPEG_HIJACK_WAITMENU, NULL );
		    return;
		}

	    default :				/* Ignore input. */
	    	break;
	}
    }
}

/*********************************************************************
 * Entrance function
 *********************************************************************/
int main( int nargs, char** args )
{
    int fd;
    int rc;
    int menu_index;

    /* I'm not sure what this line is for. Maybe for full screen graphics? */
    /* unsigned char screen[EMPEG_SCREEN_BYTES] = { 0 }; */

    /* The name that shows up in the Hijack menu. */
    char *menu[] = { "Launcher", NULL };

    daemon(0, 0); /* Change to root dir and send output to /dev/null. */

    fd = open( "/dev/display", O_RDWR );

    if( !fd )
    {
	perror( "Couldn't connect to display!\n" );
	return 0;
    }

    /* Loop while waiting to be chosen from the menu... */
    while( 1 )
    {
	/* Wait for menu... */
	rc = ioctl( fd, EMPEG_HIJACK_WAITMENU, menu ); 
	if( rc < 0 )
	{
	    perror( "Couldn't connect to Hijack menu!\n" );
	    close( fd );

	    return 0;
	}
	else
	    menu_index = rc;

	/* The claw is my master! I have been chosen! */
	show_launcher( fd, fullscreen );
    }      
}
